@@ -11,22 +11,25 @@
 ########################################################################
 
 """Here is defined the AttributeSet class."""
+from __future__ import absolute_import
 
 import re
 import sys
 import warnings
-import pickle
+import six.moves.cPickle
 import numpy
 
-from tables import hdf5extension
-from tables.utils import SizeType
-from tables.registry import class_name_dict
-from tables.exceptions import ClosedNodeError, PerformanceWarning
-from tables.path import check_name_validity
-from tables.undoredo import attr_to_shadow
-from tables.filters import Filters
+from . import hdf5extension
+from .utils import SizeType
+from .registry import class_name_dict
+from .exceptions import ClosedNodeError, PerformanceWarning
+from .path import check_attribute_name
+from .undoredo import attr_to_shadow
+from .filters import Filters
+
+from six.moves import map
+import six
 
-from tables._past import previous_api
 
 # System attributes
 SYS_ATTRS = ["CLASS", "VERSION", "TITLE", "NROWS", "EXTDIM",
@@ -69,6 +72,7 @@ def issysattrname(name):
         return False
 
 
+@six.python_2_unicode_compatible
 class AttributeSet(hdf5extension.AttributeSet, object):
     """Container for the HDF5 attributes of a Node.
 
@@ -147,9 +151,9 @@ class AttributeSet(hdf5extension.AttributeSet, object):
 
     .. rubric:: Notes on AttributeSet methods
 
-    Note that this class overrides the __getattr__(), __setattr__() and
-    __delattr__() special methods.  This allows you to read, assign or
-    delete attributes on disk by just using the next constructs::
+    Note that this class overrides the __getattr__(), __setattr__(),
+    __delattr__() and __dir__() special methods.  This allows you to
+    read, assign or delete attributes on disk by just using the next constructs::
 
         leaf.attrs.myattr = 'str attr'    # set a string (native support)
         leaf.attrs.myattr2 = 3            # set an integer (native support)
@@ -161,11 +165,15 @@ class AttributeSet(hdf5extension.AttributeSet, object):
     __delitem__() methods are available, so you may write things like
     this::
 
-        for name in :attr:`Node._v_attrs`._f_list():
-            print("name: %s, value: %s" % (name, :attr:`Node._v_attrs`[name]))
+        for name in node._v_attrs._f_list():
+            print("name: %s, value: %s" % (name, node._v_attrs[name]))
 
     Use whatever idiom you prefer to access the attributes.
 
+    Finally, on interactive python sessions you may get autocompletions of
+    attributes named as *valid python identifiers* by pressing the  `[Tab]`
+    key, or to use the dir() global function.
+
     If an attribute is set on a target node that already has a large
     number of attributes, a PerformanceWarning will be issued.
 
@@ -193,9 +201,10 @@ class AttributeSet(hdf5extension.AttributeSet, object):
     def _g_getnode(self):
         return self._v__nodefile._get_node(self._v__nodepath)
 
-    _v_node = property(_g_getnode, None, None,
-                       "The :class:`Node` instance this attribute set is "
-                       "associated with.")
+    @property
+    def _v_node(self):
+        """The :class:`Node` instance this attribute set is associated with."""
+        return self._g_getnode()
 
     def __init__(self, node):
         """Create the basic structures to keep the attribute information.
@@ -213,14 +222,14 @@ class AttributeSet(hdf5extension.AttributeSet, object):
         if not node._v_isopen:
             raise ClosedNodeError("the node for attribute set is closed")
 
-        mydict = self.__dict__
+        dict_ = self.__dict__
 
         self._g_new(node)
-        mydict["_v__nodefile"] = node._v_file
-        mydict["_v__nodepath"] = node._v_pathname
-        mydict["_v_attrnames"] = self._g_list_attr(node)
+        dict_["_v__nodefile"] = node._v_file
+        dict_["_v__nodepath"] = node._v_pathname
+        dict_["_v_attrnames"] = self._g_list_attr(node)
         # The list of unimplemented attribute names
-        mydict["_v_unimplemented"] = []
+        dict_["_v_unimplemented"] = []
 
         # Get the file version format. This is an optimization
         # in order to avoid accessing it too much.
@@ -233,10 +242,10 @@ class AttributeSet(hdf5extension.AttributeSet, object):
                 parsed_version = None
             else:
                 parsed_version = tuple(map(int, format_version.split('.')))
-        mydict["_v__format_version"] = parsed_version
+        dict_["_v__format_version"] = parsed_version
         # Split the attribute list in system and user lists
-        mydict["_v_attrnamessys"] = []
-        mydict["_v_attrnamesuser"] = []
+        dict_["_v_attrnamessys"] = []
+        dict_["_v_attrnamesuser"] = []
         for attr in self._v_attrnames:
             # put the attributes on the local dictionary to allow
             # tab-completion
@@ -254,13 +263,12 @@ class AttributeSet(hdf5extension.AttributeSet, object):
     def _g_update_node_location(self, node):
         """Updates the location information about the associated `node`."""
 
-        myDict = self.__dict__
-        myDict['_v__nodefile'] = node._v_file
-        myDict['_v__nodepath'] = node._v_pathname
+        dict_ = self.__dict__
+        dict_['_v__nodefile'] = node._v_file
+        dict_['_v__nodepath'] = node._v_pathname
         # hdf5extension operations:
         self._g_new(node)
 
-    _g_updateNodeLocation = previous_api(_g_update_node_location)
 
     def _f_list(self, attrset='user'):
         """Get a list of attribute names.
@@ -279,6 +287,15 @@ class AttributeSet(hdf5extension.AttributeSet, object):
         elif attrset == "all":
             return self._v_attrnames[:]
 
+    def __dir__(self):
+        """Autocomplete only children named as valid python identifiers.
+
+        Only PY3 supports this special method.
+        """
+        return list(set(c for c in
+                    super(AttributeSet, self).__dir__() + self._v_attrnames
+                    if c.isidentifier()))
+
     def __getattr__(self, name):
         """Get the attribute named "name"."""
 
@@ -310,24 +327,40 @@ class AttributeSet(hdf5extension.AttributeSet, object):
             # This format was used during the first 1.2 releases, just
             # for string defaults.
             try:
-                retval = pickle.loads(value)
+                retval = six.moves.cPickle.loads(value)
                 retval = numpy.array(retval)
             except ImportError:
                 retval = None  # signal error avoiding exception
-        elif maybe_pickled and name == 'FILTERS' and format_version < (2, 0):
+        elif maybe_pickled and name == 'FILTERS' and format_version is not None and format_version < (2, 0):
             # This is a big hack, but we don't have other way to recognize
             # pickled filters of PyTables 1.x files.
             value = _old_filters_re.sub(_new_filters_sub, value, 1)
-            retval = pickle.loads(value)  # pass unpickling errors through
+            retval = six.moves.cPickle.loads(value)  # pass unpickling errors through
         elif maybe_pickled:
             try:
-                retval = pickle.loads(value)
+                retval = six.moves.cPickle.loads(value)
             # except cPickle.UnpicklingError:
             # It seems that pickle may raise other errors than UnpicklingError
             # Perhaps it would be better just an "except:" clause?
             # except (cPickle.UnpicklingError, ImportError):
             # Definitely (see SF bug #1254636)
+            except UnicodeDecodeError:
+                # Object maybe pickled on python 2 and unpickled on python 3.
+                # encoding='bytes' was added in python 3.4 to resolve this.
+                # However 'bytes' mangles class attributes as they are
+                # unplicked as bytestrings. Hence try 'latin1' first.
+                # Ref: http://bugs.python.org/issue6784
+                try:
+                    retval = six.moves.cPickle.loads(value, encoding='latin1')
+                except TypeError:
+                    try:
+                        retval = six.moves.cPickle.loads(value, encoding='bytes')
+                    except:
+                        retval = value
+                except:
+                    retval = value
             except:
+                # catch other unpickling errors:
                 # ivb (2005-09-07): It is too hard to tell
                 # whether the unpickling failed
                 # because of the string not being a pickle one at all,
@@ -338,11 +371,17 @@ class AttributeSet(hdf5extension.AttributeSet, object):
                 # explaining how the user can tell where the problem was.
                 retval = value
             # Additional check for allowing a workaround for #307
-            if isinstance(retval, str) and retval == '':
+            if isinstance(retval, six.text_type) and retval == u'':
                 retval = numpy.array(retval)[()]
-        elif name == 'FILTERS' and format_version >= (2, 0):
+        elif name == 'FILTERS' and format_version is not None and format_version >= (2, 0):
             retval = Filters._unpack(value)
-        elif (issysattrname(name) and isinstance(value, (bytes, str)) and
+        elif name == 'TITLE' and not isinstance(value, str):
+            if sys.version_info[0] < 3:
+                # unicode is OK for TITLE
+                retval = value
+            else:
+                retval = value.decode('utf-8')
+        elif (issysattrname(name) and isinstance(value, (bytes, six.text_type)) and
               not isinstance(value, str) and not _field_fill_re.match(name)):
             # system attributes should always be str
             if sys.version_info[0] < 3:
@@ -378,18 +417,18 @@ class AttributeSet(hdf5extension.AttributeSet, object):
             elif name == "NROWS":
                 stvalue = numpy.array(value, dtype=SizeType)
                 value = stvalue[()]
-            elif name == "FILTERS" and self._v__format_version >= (2, 0):
+            elif name == "FILTERS" and self._v__format_version is not None and self._v__format_version >= (2, 0):
                 stvalue = value._pack()
                 # value will remain as a Filters instance here
         # Convert value from a Python scalar into a NumPy scalar
         # (only in case it has not been converted yet)
         # Fixes ticket #59
         if (stvalue is value and
-                type(value) in (bool, bytes, int, float, complex, str,
+                type(value) in (bool, bytes, int, float, complex, six.text_type,
                                 numpy.unicode_)):
             # Additional check for allowing a workaround for #307
-            if value == '':
-                stvalue = numpy.array('')
+            if isinstance(value, six.text_type) and len(value) == 0:
+                stvalue = numpy.array(u'')
             else:
                 stvalue = numpy.array(value)
             value = stvalue[()]
@@ -429,48 +468,46 @@ class AttributeSet(hdf5extension.AttributeSet, object):
 
         """
 
-        nodeFile = self._v__nodefile
+        nodefile = self._v__nodefile
         attrnames = self._v_attrnames
 
         # Check for name validity
-        check_name_validity(name)
+        check_attribute_name(name)
 
-        nodeFile._check_writable()
+        nodefile._check_writable()
 
         # Check if there are too many attributes.
-        maxNodeAttrs = nodeFile.params['MAX_NODE_ATTRS']
-        if len(attrnames) >= maxNodeAttrs:
+        max_node_attrs = nodefile.params['MAX_NODE_ATTRS']
+        if len(attrnames) >= max_node_attrs:
             warnings.warn("""\
 node ``%s`` is exceeding the recommended maximum number of attributes (%d);\
 be ready to see PyTables asking for *lots* of memory and possibly slow I/O"""
-                          % (self._v__nodepath, maxNodeAttrs),
+                          % (self._v__nodepath, max_node_attrs),
                           PerformanceWarning)
 
-        undoEnabled = nodeFile.is_undo_enabled()
+        undo_enabled = nodefile.is_undo_enabled()
         # Log old attribute removal (if any).
-        if undoEnabled and (name in attrnames):
+        if undo_enabled and (name in attrnames):
             self._g_del_and_log(name)
 
         # Set the attribute.
         self._g__setattr(name, value)
 
         # Log new attribute addition.
-        if undoEnabled:
+        if undo_enabled:
             self._g_log_add(name)
 
     def _g_log_add(self, name):
         self._v__nodefile._log('ADDATTR', self._v__nodepath, name)
 
-    _g_logAdd = previous_api(_g_log_add)
 
     def _g_del_and_log(self, name):
-        nodeFile = self._v__nodefile
-        nodePathname = self._v__nodepath
+        nodefile = self._v__nodefile
+        node_pathname = self._v__nodepath
         # Log *before* moving to use the right shadow name.
-        nodeFile._log('DELATTR', nodePathname, name)
-        attr_to_shadow(nodeFile, nodePathname, name)
+        nodefile._log('DELATTR', node_pathname, name)
+        attr_to_shadow(nodefile, node_pathname, name)
 
-    _g_delAndLog = previous_api(_g_del_and_log)
 
     def _g__delattr(self, name):
         """Delete a PyTables attribute.
@@ -504,7 +541,7 @@ be ready to see PyTables asking for *lots* of memory and possibly slow I/O"""
 
         """
 
-        nodeFile = self._v__nodefile
+        nodefile = self._v__nodefile
 
         # Check if attribute exists
         if name not in self._v_attrnames:
@@ -512,10 +549,10 @@ be ready to see PyTables asking for *lots* of memory and possibly slow I/O"""
                 "Attribute ('%s') does not exist in node '%s'"
                 % (name, self._v__nodepath))
 
-        nodeFile._check_writable()
+        nodefile._check_writable()
 
         # Remove the PyTables attribute or move it to shadow.
-        if nodeFile.is_undo_enabled():
+        if nodefile.is_undo_enabled():
             self._g_del_and_log(name)
         else:
             self._g__delattr(name)
@@ -669,12 +706,10 @@ class NotLoggedAttributeSet(AttributeSet):
     def _g_log_add(self, name):
         pass
 
-    _g_logAdd = previous_api(_g_log_add)
 
     def _g_del_and_log(self, name):
         self._g__delattr(name)
 
-    _g_delAndLog = previous_api(_g_del_and_log)
 
 ## Local Variables:
 ## mode: python
